/*-
 * Copyright (C) @BABOLO  2002 Dec 04
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2006 Feb 21\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: cmd_trans.c,v 1.8 2007/11/11 11:46:28 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sysexits.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <poll.h>
#include <mife.h>
#include <err.h>
#include <libpq-fe.h>
#include <babolo/BLINflag.h>
#include "pgoblin.h"
#include "pgob.h"

int
pgoblin_fork(pgoblin_con *options, int a[]) {
    pid_t pid;
    int ex = EX_OK;

    ifBLIN_QV3(options->flags) fprintf(stderr, "+pgoblin_fork\n");
    if  ((pid = fork()) < 0) {
        ex = EX_OSERR;
    }
    ifBLIN_QV3(options->flags) fprintf( stderr, "-pgoblin_fork %d\n", ex);
    if  (pid) exit(0);
    return(ex);
}

int
pgoblin_begin(pgoblin_con *options, int a[]) {
    PGresult *res;

    ifBLIN_QV3(options->flags) fprintf(stderr, "+pgoblin_begin\n");
    if  (!(R_CONN.intran)) {
        PQEXEC(res, "BEGIN", PGRES_COMMAND_OK, -1);
        PQclear(res);
        R_CONN.intran++;
    }
    ifBLIN_QV3(options->flags) fprintf( stderr, "-pgoblin_begin %d\n", 0);
    return(0);
}

int
pgoblin_end(pgoblin_con *options, int a[]) {
    PGresult *res;

    ifBLIN_QV3(options->flags) fprintf(stderr, "+pgoblin_end\n");
    if  ((R_CONN.intran)) {
        PQEXEC(res, "END", PGRES_COMMAND_OK, -1);
        PQclear(res);
        R_CONN.intran--;
    }
    ifBLIN_QV3(options->flags) fprintf( stderr, "-pgoblin_end %d\n", 0);
    return(0);
}

int
pgoblin_listen(pgoblin_con *options, int a[]) {
    int       sock;
    struct pollfd pfd;
    int timeout, i, e, ex = EX_OK;
    int64_t tout, unixtimeto;
    PGnotify *notify;
    struct timeval t;
    union {
        double  ftime;
        int64_t itime;
        u_int32_t d[2];
    } dbtime;
    union {
        u_int64_t p;
        u_int32_t d[2];
    } pmatsemit;

    ifBLIN_QV3(options->flags) fprintf( stderr, "+pgoblin_listen %d %d %d =%s~\n"
                                      , a[PGO_PCTL], a[PGO_PCON], a[PGO_PIN]
                                      , (char*)IO_CTL.text
                                      );
    if  ((sock = PQsocket(R_CONN.conn)) < 0) {
        ex = EX_UNAVAILABLE;
        errno = EINVAL;
    } else if (  PQgetisnull(IO_IN.pq, 0, 0)
              || (PQgetlength(IO_IN.pq, 0, 0) < 8)
              ) {
        ex = EX_DATAERR;
        errno = EINVAL;
    } else {
        bcopy(PQgetvalue(IO_IN.pq, 0, 0), &pmatsemit, sizeof(pmatsemit));
        dbtime.d[1] = ntohl(pmatsemit.d[0]);
        dbtime.d[0] = ntohl(pmatsemit.d[1]);
        if  (R_CONN.flags & PGOBLIN_FLOATTIME) {
            unixtimeto = (dbtime.ftime * 1000000.0) + (int64_t)(10957LL * 86400LL * 1000000LL);
        } else if (R_CONN.flags & PGOBLIN_INT64TIME) {
            unixtimeto = dbtime.itime + (int64_t)(10957LL * 86400LL * 1000000LL);
        } else {
            unixtimeto = 0;
            ex = EX_DATAERR;
            errno = EINVAL;
        }
        bzero(&pfd, sizeof(pfd));
        pfd.fd = sock;
        pfd.events = POLLIN | POLLRDNORM | POLLRDBAND | POLLPRI;
        for (e = 1; e && !ex;) {
            if  (gettimeofday(&t, NULL)) {
                ex = EX_OSERR;
                break;
            }
            /* In PostgreSQL timestamp is in float seconds
             * or in integer 1/1000000 seconds from 2000-01-01
             */
            tout = unixtimeto - ((int64_t)1000000LL * t.tv_sec + t.tv_usec);
            if  (tout <= 0) break;
            else if ((tout / 1000) > options->maxtmout) {
                timeout = options->maxtmout;
                unixtimeto = ((int64_t)1000LL * options->maxtmout)
                           + ((int64_t)1000000LL * t.tv_sec)
                           + t.tv_usec
                           ;
            } else timeout = tout / 1000;
            ifBLIN_QV4(options->flags) fprintf( stderr, "LISTEN=%s~ fd %d for %d ms\n"
                                              , (char*)IO_CTL.text, sock, timeout
                                              );
            if  ((i = poll(&pfd, 1, timeout)) < 0) {
                ex = EX_OSERR;
                break;
            }
            PQconsumeInput(R_CONN.conn);
            while ( (notify = PQnotifies(R_CONN.conn))) {
                ifBLIN_QV4(options->flags)
                    fprintf(stderr, "Received NOTIFY=%s~\n", notify->relname);
                if  (!strcmp(notify->relname, IO_CTL.text)) e = 0;
                PQfreemem(notify);
    }   }   }
    return(ex);
}

int
pgoblin_exec(pgoblin_con *options, int a[]) {
    char *argv[256];
    int i;
    int ex = EX_OK;

    ifBLIN_QV3(options->flags) fprintf(stderr, "+pgoblin_exec\n");
    if  ((dup2(IO_OUT.onu, fileno(stdout))) < 0) { /* XXXX onu!!! */
        ex = EX_OSERR;
        ifBLIN_QV1(options->flags) warn("dup2");
    } else if ((dup2(IO_IN.mife->fd, fileno(stdin))) < 0) { /* XXXX purge */
        ex = EX_OSERR;
        ifBLIN_QV1(options->flags) warn("dup2");
    } else {
        if  (IO_OUT.onu != fileno(stdout)) close(IO_OUT.onu);
        if  (IO_IN.mife->fd != fileno(stdin)) close(IO_IN.mife->fd);
        if  (!a[PGO_PCTL]) {
            ifBLIN_QV4(options->flags) fprintf(stderr, "exec from 0\n");
            for (i = 0; ; i++) {
               if  (i && options->pgm->code[options->ppoint + i].cmd != PGOBLIN_cont) break;
               if  (!options->pgm->code[options->ppoint + i].lit) break;
               argv[i] = options->pgm->code[options->ppoint + i].lit;
            }
            argv[i] = NULL;
            execve(argv[0], argv, options->envp);
            ex = EX_OSERR;
            ifBLIN_QV1(options->flags) warn("execve");
        } else if (!PQntuples(IO_CTL.pq)) {
            ifBLIN_QV4(options->flags) fprintf(stderr, "null exec from %d\n", a[PGO_PCTL]);
        } else if (PQntuples(IO_CTL.pq) == 1) {
            int c;

            ifBLIN_QV4(options->flags) fprintf(stderr, "1 exec from %d\n", a[PGO_PCTL]);
            for (i = 0, c = 0; c < PQnfields(IO_CTL.pq); c++) {
                if  (!PQgetisnull(IO_CTL.pq, 0, c))
                    argv[i++] = PQgetvalue(IO_CTL.pq, 0, c);
            }
            if  (i) {
                argv[i] = NULL;
                execve(argv[0], argv, options->envp);
                ex = EX_OSERR;
                ifBLIN_QV1(options->flags) warn("execve");
            }
        } else {
            int k, c;

            ifBLIN_QV4(options->flags) fprintf(stderr, "multi exec from %d\n", a[PGO_PCTL]);
            for (k = 0; k < PQntuples(IO_CTL.pq); k++) {
                for (i = 0, c = 0; c < PQnfields(IO_CTL.pq); c++) {
                    if  (!PQgetisnull(IO_CTL.pq, k, c))
                        argv[i++] = PQgetvalue(IO_CTL.pq, k, c);
                }
                if  (i) {
                    argv[i] = NULL;
                    if  ((ex = fork()) < 0) {
                        ifBLIN_QV1(options->flags) warn("fork #3");
                        return(EX_OSERR);
                    } else if (!ex) {
                        execve(argv[0], argv, options->envp);
                        ex = EX_OSERR;
                        ifBLIN_QV1(options->flags) warn("execve");
                    } else {
                        /* Lost PID XXXX */
                        ex = 0;
            }   }   }
            exit(ex);
    }   }
    ifBLIN_QV3(options->flags) fprintf( stderr, "-pgoblin_exec %d\n", ex);
    return(ex);
}
