/*-
 * Copyright (C) @BABOLO  2002 Dec 23
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Dec 23\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: execute.c,v 1.17 2007/11/10 22:00:36 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/wait.h>
#include <sysexits.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <mife.h>
#include <err.h>
#include <libpq-fe.h>
#include <babolo/BLINflag.h>
#include "pgoblin.h"
#include "pgob.h"

static void
dump_reg(pgoblin_con *options, int r) {
    u_int32_t wdr, cdr;

    fprintf( stderr, "%d(%08X): ref=%d pid=%d\n"
           , r, options->io[r].flags, options->io[r].ref, options->io[r].pid
           );
    if  (options->io[r].text) {
        if  (options->io[r].flags & PGOBLIN_BINPARM)
            fprintf(stderr, "bin=~\n", (u_int32_t)options->io[r].text);
          else fprintf(stderr, "text=%s~\n", (char*)options->io[r].text);
    }
    if  (options->io[r].mife) fprintf(stderr, "mife=%08X\n", (u_int32_t)options->io[r].mife);
    if  (options->io[r].pq) fprintf(stderr, "pq=%08X\n", (u_int32_t)options->io[r].pq);
    wdr = (u_int32_t)options->io[r].wri;
    if  (wdr == (u_int32_t)write) fprintf(stderr, "wri=write ");
    else if (wdr == (u_int32_t)mife_writ) fprintf(stderr, "wri=mife_writ ");
    else if (wdr) fprintf(stderr, "wri=%08X ", wdr);
    cdr = (u_int32_t)options->io[r].clo;
    if  (cdr == (u_int32_t)close) fprintf(stderr, "clo=close ");
    else if (cdr == (u_int32_t)mife_close) fprintf(stderr, "clo=mife_close ");
    else if (cdr) fprintf(stderr, "clo=%08X ", cdr);
    if  (wdr || cdr) fprintf(stderr, "onu=%d\n", options->io[r].onu);
}

#define Pt PGOBLIN_TEXTPARM
#define Pf PGOBLIN_FILENAME
#define Pm PGOBLIN_MIFEDES
#define Pq PGOBLIN_PQRESULT
#define Po PGOBLIN_OUTSET
#define Ps PGOBLIN_STYLE
#define Pc PGOBLIN_CONNECT

#define Fx PGOBLIN_ILLEGAL
#define Fi PGOBLIN_IGNORE
#define Fl PGOBLIN_NOLITER
#define Io PGOBLIN_ISOUT
#define Id PGOBLIN_ISCTL
#define Ii PGOBLIN_ISIN
#define Ic PGOBLIN_ISCONN
#define Is PGOBLIN_ISSTYLE

const pgoblin_c pgoblin_a[PGOBLIN_max] =
{{ Fx                 , {    0,  0,     0,  0}, NULL           , NULL}
,{                   0, {    0,  0,     0,  0}, NULL           , ""}
,{ Fi                 , {    0,  0,     0,  0}, NULL           , "!"}
,{ Fx                 , {    0,  0, Po|Pc,  0}, NULL           , "connect"}
,{       Id|Ii|Ic     , {    0, Pt,    Pc, Pq}, pgoblin_listen , "listen"}
,{ Fl|         Ic     , {    0,  0,    Pc,  0}, pgoblin_begin  , "begin"}
,{ Fl|         Ic     , {    0,  0,    Pc,  0}, pgoblin_end    , "end"}
,{       Id|   Ic     , {    0, Pt,    Pc,  0}, pgoblin_perform, "perform"}
,{    Io|Id           , {Po|Pt, Pt,     0,  0}, pgoblin_echo   , "echo"}
,{    Io|Id           , {Po|Pm, Pt,     0,  0}, pgoblin_cat    , "cat"}
,{    Io|Id|Ii        , {Po|Pm, Pq,     0, Pm}, pgoblin_exec   , "exec"}
,{ Fl                 , {    0,  0,     0,  0}, pgoblin_fork   , "fork"}
,{ Fx                 , {    0,  0,     0,  0}, NULL           , "include"}
,{       Id|Ii|Ic     , {    0, Pt,    Pc, Pm}, pgoblin_copyin , "copyin"}
,{    Io|Id|   Ic     , {Po|Pm, Pt,    Pc,  0}, pgoblin_copyout, "copyout"}
,{    Io|Id|   Ic|Is|3, {Po|Pm, Pt,    Pc, Ps}, pgoblin_binsel , "binselect"}
,{    Io|Id|   Ic|Is|3, {Po|Pm, Pt,    Pc, Ps}, pgoblin_strsel , "strselect"}
,{    Io|Id|   Ic     , {Po|Pq, Pt,    Pc,  0}, pgoblin_binpre , "binpresel"}
,{    Io|Id|   Ic     , {Po|Pq, Pt,    Pc,  0}, pgoblin_select , "select"}
,{ Fl|Io|   Ii|   Is|2, {Po|Pm,  0,    Ps, Pq}, pgoblin_display, "display"}
,{ Fl|Io              , {Po|Pm,  0,     0,  0}, pgoblin_getarg , "getarg"}
,{ Fl|Io              , {Po|Pm,  0,     0,  0}, pgoblin_getenv , "getenv"}
,{ Fl|Io              , {Po|Pm,  0,     0,  0}, pgoblin_getget , "getget"}
,{ Fl|Io|   Ii        , {Po|Pm,  0,     0, Pm}, pgoblin_getpost, "getpost"}
,{ Fl|Io              , {Po|Pm,  0,     0,  0}, pgoblin_getproc, "getproc"}
};

static void
freereg(pgoblin_con *options, int r, u_int32_t m) {
    ifBLIN_QV4(options->flags) fprintf(stderr, "+freereg %d %08X\n", r, m);
    options->io[r].ref = 0;
    options->io[r].flags &= ~PGOBLIN_REFERER;
    options->io[r].flags &= ~PGOBLIN_CTLARRAY;
    if  (m & (PGOBLIN_TEXTPARM | PGOBLIN_BINPARM | PGOBLIN_FILENAME)) {
        if  ((options->io[r].flags & PGOBLIN_FREETEXT) && (options->io[r].text))
            free(options->io[r].text);
        options->io[r].text = NULL;
        options->io[r].length = 0;
        options->io[r].flags &= ~(PGOBLIN_FREETEXT | PGOBLIN_TEXTPARM | PGOBLIN_BINPARM);
    }
    if  (m & PGOBLIN_PQRESULT) {
        if  ((options->io[r].flags & PGOBLIN_FREEPQRE) && (options->io[r].pq))
            PQclear(options->io[r].pq);
        options->io[r].pq = NULL;
        options->io[r].flags &= ~(PGOBLIN_FREEPQRE | PGOBLIN_PQRESULT);
    }
    if  (m & PGOBLIN_MIFEDES) {
        if  ((options->io[r].flags & PGOBLIN_FREEMIFE) && (options->io[r].mife))
            mife_close(options->io[r].mife);
        options->io[r].mife = NULL;
        options->io[r].offset = 0;
        options->io[r].pid = 0;
        options->io[r].flags &= ~(PGOBLIN_FREEMIFE | PGOBLIN_MIFEDES);
    }
    ifBLIN_QV4(options->flags) fprintf(stderr, "-freereg\n");
}

#define CC 0x000F /* price */
#define Cx 0x8000 /* impossible */
#define Cq 0x4000 /* read to text */
#define Co 0x2000 /* open file name */
#define Cr 0x1000 /* read file */
#define Cn 0x0800 /* temp file name */
#define Cw 0x0400 /* write file */
#define Cm 0x0200 /* do mife from text */
#define Ca 0x0100 /* throw pq to file */
#define Cb 0x0080 /* throw mife to file */

static const u_int16_t score[4][4] =
      /*      Pt       Pm       Pq       Pf   From */
{/* Pt */{       0, Cr   |2, Cq   |5, Co|Cr|3}
,/* Pm */{    Cm|1,       0, Cq|Cm|6, Co   |1}
,/* Pq */{ Cx  |15, Cx  |15,       0, Cx  |15}
,/* Pf */{ Cn|Cw|7, Cn|Cb|9, Cn|Ca|8,       0}
};

static int
coercion(pgoblin_con *options, int r, int c, pgoblin_command cmd) {
    int i, j;
    u_int32_t price;

    /* if pgoblin_a[cmd].ch[c] & PGOBLIN_OUTSET ? */
    if  (r) {
        if  (  ((options->io[r].flags & PGOBLIN_ANYTEXT) && !options->io[r].text)
            || ((options->io[r].flags & PGOBLIN_MIFEDES) && !options->io[r].mife)
            || ((options->io[r].flags & PGOBLIN_PQRESULT) && !options->io[r].pq)
            ) {
            ifBLIN_QV1(options->flags) warnx("No info for in=%d f=%08X", r, options->io[r].flags);
            return(-1);
        }
        price = Cx | 15;
        ifBLIN_QV4(options->flags) fprintf( stderr, "r=%d ch=%02X io=%08X%s%s%s\n"
                                          , r, pgoblin_a[cmd].ch[c], options->io[r].flags
                                          , options->io[r].text ? " text" : ""
                                          , options->io[r].mife ? " mife" : ""
                                          , options->io[r].pq   ? " pq"   : ""
                                          );
        for (i = 0; i < 3; i++) { /* To */
            if  (  ((i == 0) && (pgoblin_a[cmd].ch[c] & PGOBLIN_ANYTEXT))
                || ((i == 1) && (pgoblin_a[cmd].ch[c] & PGOBLIN_MIFEDES))
                || ((i == 2) && (pgoblin_a[cmd].ch[c] & PGOBLIN_PQRESULT))
                ) {
                for (j = 0; j < 3; j++) { /* From */
                    if  (  (  ((j == 0) && (options->io[r].text))
                           || ((j == 1) && (options->io[r].mife))
                           || ((j == 2) && (options->io[r].pq))
                           )
                        && (score[i][j] & CC) < (price & CC)
                        ) {
                        price = score[i][j];
                        ifBLIN_QV4(options->flags)
                            fprintf(stderr, "i=%d j=%d %04X\n", i, j, score[i][j]);
        }   }   }   }
        if  (price & Cx) {
                ifBLIN_QV1(options->flags)
                    warnx("No way @%d for in=%d f=%08X", options->ppoint, r, options->io[r].flags);
                return(-1);
        }
        if  (price & Cq) {
            ssize_t k, c, e, l;
            char *s;

            for (k = 0, e = 0; k < PQntuples(options->io[r].pq) && e >= 0; k++) {
                for (c = 0; c < PQnfields(options->io[r].pq) && e >= 0; c++) {
                    if  (!PQgetisnull(options->io[r].pq, k, c))
                        e += PQgetlength(options->io[r].pq, k, c);
            }   }
            freereg(options, r, PGOBLIN_ANYTEXT);
            if  (!(s = malloc(e + 1))) {
                ifBLIN_QV1(options->flags) warnx("No mem");
                return(-1);
            }
            options->io[r].text = s;
            options->io[r].length = e;
            options->io[r].flags |= PGOBLIN_FREETEXT;
            for (k = 0, e = 0; k < PQntuples(options->io[r].pq) && e >= 0; k++) {
                for (c = 0; c < PQnfields(options->io[r].pq) && e >= 0; c++) {
                    if  (!PQgetisnull(options->io[r].pq, k, c)) {
                        l = PQgetlength(options->io[r].pq, k, c);
                        bcopy(PQgetvalue(options->io[r].pq, k, c), s, l);
                        s += l;
            }   }   }
            *s = '\0';
        }
        if  (price & Co) {
            freereg(options, r, PGOBLIN_MIFEDES);
            /* XXXX */
        }
        if  (price & Cr) {
            freereg(options, r, PGOBLIN_ANYTEXT);
            options->io[r].length = mife_read(options->io[r].mife, 0, 0);
            if  (!(options->io[r].length)) {
                ifBLIN_QV1(options->flags) warnx("mife_read");
                return(-1);
            }
            options->io[r].text = mife_get(options->io[r].mife, 0);
        }
        if  (price & Cn) /* XXXX */;
        if  (price & Cw) /* XXXX */;
        if  (price & Cm) {
            freereg(options, r, PGOBLIN_MIFEDES);
            if  (options->io[r].flags & PGOBLIN_FREETEXT) {
                options->io[r].mife = mife_opem( options->io[r].length ? 0 : MIFE_FULL
                                               , options->io[r].text
                                               , options->io[r].length
                                               ? options->io[r].length
                                               : strlen(options->io[r].text)
                                               );
            } else {
                options->io[r].mife = mife_opes( MIFE_FULL
                                               , options->io[r].text
                                               , options->io[r].length
                                               ? options->io[r].length
                                               : strlen(options->io[r].text)
                                               );

                options->io[r].flags |= PGOBLIN_FREETEXT;
        }   }
        if  (price & Ca) /* XXXX */;
        if  (price & Cb) /* XXXX */;
    } else {
        ifBLIN_QV1(options->flags) warnx("r=0");
        return(-1);
    }
    return(0);
}

static void
pgoblin_io_c(pgoblin_con *options, int r, u_int32_t m) {
    ifBLIN_QV4(options->flags) fprintf(stderr, "+pgoblin_io_c %d %08X\n", r, m);
    if  (r || (m & PGOBLIN_FORCE0)) {
        if  (m & PGOBLIN_OUTSET) {
            if  (options->io[r].clo) {
                if  (options->io[r].flags & PGOBLIN_FREEOUTS) {
                    options->io[r].clo(options->io[r].onu);
                }
                options->io[r].clo = NULL;
            }
            if  (options->io[r].wri) options->io[r].wri = NULL;
            m = PGOBLIN_TEXTPARM | PGOBLIN_MIFEDES | PGOBLIN_PQRESULT;
        }
        if  (m & PGOBLIN_PQRESULT) {
            if  (options->io[r].pq) {
                if  (options->io[r].flags & PGOBLIN_FREEPQRE) {
                    PQclear(options->io[r].pq);
                }
                options->io[r].pq = NULL;
        }   }
        if  (m & PGOBLIN_MIFEDES) {
            if  (options->io[r].mife) {
                if  (options->io[r].flags & PGOBLIN_FREEMIFE) {
                    mife_close(options->io[r].mife);
                }
            options->io[r].mife = NULL;
        }   }
        if  (m & PGOBLIN_TEXTPARM) {
            if  (options->io[r].text) {
                if  (options->io[r].flags & PGOBLIN_FREETEXT) {
                    free(options->io[r].text);
                }
            options->io[r].text = NULL;
    }   }   }
    ifBLIN_QV4(options->flags) fprintf(stderr, "-pgoblin_io_c\n");
}

int
/***************************************************
 *                                                 */
pgoblin_execute(pgoblin_con *options) {
/*                                                 *
 ***************************************************/
    int pipa[2];
    int ex, i;

    ifBLIN_QV3(options->flags) fprintf( stderr, "+pgoblin_execute pgm@%d:%d\n"
                                      , options->pgm->curr, options->pgm->maxx
                                      );
    ifBLIN_QV5(options->flags) for (i = 0; i < PGOBLIN_REGSIZE; i++) dump_reg(options, i);
    for (options->ppoint = 0, ex = 0; !ex && options->ppoint < options->pgm->curr; options->ppoint++) {
        int a[4], efl, status;
        u_int32_t t;
        pgoblin_command cmd;

        t = options->pgm->code[options->ppoint].cmd;
        cmd = t & 0xFF;
        if  (cmd >= PGOBLIN_max || pgoblin_a[cmd].flags & PGOBLIN_ILLEGAL) {
            ifBLIN_QV1(options->flags) warnx("unknown command %d", options->ppoint);
            return(-1);
        }
        t >>= 2;
        for (i = 0; i < 4; i++) {
            t >>= 6;
            a[i] = t & 0x3F;
        }
        if  (pgoblin_a[cmd].flags & PGOBLIN_ISSTYLE) {
            /* When style come in */
        }
        if  ((pgoblin_a[cmd].flags & PGOBLIN_ISCONN) && (CH_CON & PGOBLIN_CONNECT)) {
            if  ((CH_CON & PGOBLIN_OUTSET) && (R_CONN.conn)) {
                PQfinish(R_CONN.conn);
                R_CONN.conn = NULL;
            } else if (pgoblin_shurecon(options, a)) {
                ifBLIN_QV1(options->flags) warn("pgoblin_shurecon");
                return(-1);
        }   }
        if  (pgoblin_a[cmd].flags & PGOBLIN_ISIN) {
            /* if CH_IN & PGOBLIN_OUTSET ? XXXX */
            if  (a[PGO_PIN]) {
                if  (coercion(options, a[PGO_PIN], PGO_PIN, cmd)) {
                    ifBLIN_QV1(options->flags) warn("coercion in=%d", a[PGO_PIN]);
                    return(-1);
                }
            } else {
                /* No coercion for IN = 0 XXXX */
        }   }
        if  (pgoblin_a[cmd].flags & PGOBLIN_ISCTL) {
            /* if CH_CTL & PGOBLIN_OUTSET ? XXXX */
            if  (a[PGO_PCTL]) {
                if  (coercion(options, a[PGO_PCTL], PGO_PCTL, cmd)) {
                    ifBLIN_QV1(options->flags) warn("coercion in=%d", a[PGO_PCTL]);
                    return(-1);
                }
            } else {
                if  ((CH_CTL & PGOBLIN_TEXTPARM)) {
                    freereg(options, 0, PGOBLIN_TEXTPARM);
                    IO_RG0.text = options->pgm->code[options->ppoint].lit;
                    IO_RG0.flags |= PGOBLIN_TEXTPARM;
                }
                if  ((CH_CTL & PGOBLIN_PQRESULT)) {
                    freereg(options, 0, PGOBLIN_PQRESULT);
                    IO_RG0.flags |= PGOBLIN_CTLARRAY;
                }
        }   }
        ifBLIN_QV5(options->flags)
            for (i = 0; i < 4; i++) if (pgoblin_a[cmd].ch[i]) dump_reg(options, a[i]);
        efl = (  (pgoblin_a[cmd].flags & PGOBLIN_ISOUT)
              && (CH_OUT & PGOBLIN_OUTSET)
              && (CH_OUT & PGOBLIN_MIFEDES)
              );
        if  (efl && (a[PGO_POUT])) {
            pgoblin_io_c(options, a[PGO_POUT], PGOBLIN_OUTSET);
            IO_OUT.flags = PGOBLIN_OUTSET;
            if  ((ex = pipe(pipa))) {
                ifBLIN_QV1(options->flags) warn("pipe");
                return(EX_OSERR);
            }
            ifBLIN_QV4(options->flags)
                fprintf( stderr, "FORK %08X %s %d%d%d%d\n"
                       , options->pgm->code[options->ppoint].cmd
                       , pgoblin_a[cmd].str
                       , a[0], a[1], a[2], a[3]
                       );
            ex = fork();
            if  (ex < 0) {
                ifBLIN_QV1(options->flags) warn("fork #1");
                return(EX_OSERR);
            }
            if  (!ex) {
                IO_OUT.onu = pipa[1];
                IO_OUT.wri = mife_writ;
                IO_OUT.clo = close;
                if  (pgoblin_a[cmd].car && (ex = (*(pgoblin_a[cmd].car))(options, a))) {
                    ifBLIN_QV1(options->flags) warn("exec err");
                }
                exit(ex);
            }
            close(pipa[1]);
            IO_OUT.pid = ex;
            ex = 0;
            if  (options->flags & PGOBLIN_OLDSTYLE && a[PGO_POUT] == 1) {
                if  (IO_RG0.mife) mife_close(IO_RG0.mife);
                /* pid XXXX */
                if  (!(IO_RG0.mife = mife_opef(MIFE_FULL, pipa[0]))) {
                    warn("execute");
                    ex = EX_NOINPUT;
                }
            } else {
                if  (!(IO_OUT.mife = mife_opef(MIFE_FULL, pipa[0]))) {
                    warn("execute");
                    ex = EX_NOINPUT;
            }   }
        } else if (efl && (cmd == PGOBLIN_exec)) {
            ifBLIN_QV4(options->flags)
                fprintf( stderr, "EXEC %08X %s %d%d%d%d\n"
                       , options->pgm->code[options->ppoint].cmd
                       , pgoblin_a[cmd].str
                       , a[0], a[1], a[2], a[3]
                       );
            ex = fork();
            if  (ex < 0) {
                ifBLIN_QV1(options->flags) warn("fork #2");
                return(EX_OSERR);
            }
            if  (!ex) {
                if  (pgoblin_a[cmd].car && (ex = (*(pgoblin_a[cmd].car))(options, a))) {
                    ifBLIN_QV1(options->flags) warn("exec err");
                }
                exit(ex);
            }
            IO_OUT.pid = ex;
            do {
               if  ((ex = waitpid(IO_OUT.pid, &status, 0)) < 0) {
                   ifBLIN_QV1(options->flags) warn("wait err");
               } else if (IO_OUT.pid != ex) continue;
            } while (!WIFEXITED(status));
            if  (ex > 0) ex = WEXITSTATUS(status);
            IO_OUT.pid = 0;
        } else {
            ifBLIN_QV4(options->flags)
                fprintf( stderr, "DO %08X %s %d%d%d%d\n"
                       , options->pgm->code[options->ppoint].cmd
                       , pgoblin_a[cmd].str
                       , a[0], a[1], a[2], a[3]
                       );
            if  (pgoblin_a[cmd].car) ex = (*(pgoblin_a[cmd].car))(options, a);
        }
        if  (!ex && (pgoblin_a[cmd].flags & PGOBLIN_ISOUT) && (a[PGO_POUT] == 0)) {
            if  ((CH_OUT & PGOBLIN_TEXTPARM) && IO_OUT.text) {
                ex = IO_OUT.wri( IO_OUT.onu
                               , IO_OUT.text
                               , IO_OUT.length ? IO_OUT.length : strlen(IO_OUT.text)
                               );
                freereg(options, a[PGO_POUT], PGOBLIN_TEXTPARM);
	        if  (ex < 0) {
    	            ifBLIN_QV1(options->flags) warn("fin write");
                } else {
    	            ex = 0;
            }   }
            if  ((CH_OUT & PGOBLIN_PQRESULT) && IO_OUT.pq) {
                int k, c;

                for (k = 0, ex = 0; k < PQntuples(IO_OUT.pq) && ex >= 0; k++) {
                    for (c = 0; c < PQnfields(IO_OUT.pq) && ex >= 0; c++) {
                        if  (!PQgetisnull(IO_OUT.pq, k, c))
                            ex = IO_OUT.wri( IO_OUT.onu
                                           , PQgetvalue(IO_OUT.pq, k, c)
                                           , PQgetlength(IO_OUT.pq, k, c)
                                           );
                }   }
                freereg(options, a[PGO_POUT], PGOBLIN_PQRESULT);
	        if  (ex < 0) {
    	            ifBLIN_QV1(options->flags) warn("fin write");
                } else {
    	            ex = 0;
            }   }
            if  ((CH_OUT & IO_OUT.flags & PGOBLIN_FILENAME) && IO_OUT.text) {
                /* XXXX */
    }   }   }
    for (i = 0; i < PGOBLIN_REGSIZE; i++) {
        pgoblin_io_c(options, i, PGOBLIN_OUTSET | PGOBLIN_FORCE0);
        if  (options->conn[i].conn) {
            PQfinish(options->conn[i].conn);
            options->conn[i].conn = NULL;
        }
        if  (options->style[i]);
    }
    ifBLIN_QV3(options->flags) fprintf(stderr, "-pgoblin_execute %d\n", ex);
    return(ex);
};
