/*-
 * Copyright (C) @BABOLO  2002 Dec 04
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Dec 04\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: shurecon.c,v 1.13 2007/11/10 22:00:36 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <mife.h>
#include <err.h>
#include <libpq-fe.h>
#include <babolo/BLINflag.h>
#include "pgoblin.h"
#include "pgob.h"

#define PRINTNOTNULL(A) (!(A) ? "" : (A))
#define PRINTANDNULL(A) PRINTNOTNULL(A), ((A) ? "~" : "")
#define PQSTATNDNULL(A) PRINTANDNULL(PQparameterStatus(R_CONN.conn, (A)))
#define ALIGN8(A)  popt[i].A, (strlen(popt[i].A) < 8 ? "		" : "	")
#define ALIGN16(A) popt[i].A                                                                          \
      , ((strlen(popt[i].A) < 8) ? "		" : (strlen(popt[i].envvar) < 16) ? "	" : " ")

static void
NoticeProcessor(void *options, const char *message) {
    ifBLIN_QV1(((pgoblin_con *)options)->flags) fprintf(stderr, "%s", message);
}

int
pgoblin_shurecon(pgoblin_con *options, int a[]) {
    PQconninfoOption *popt;
    int i, ex = EX_OK;
    PGresult *res;

    if  (  R_CONN.conn != NULL
        && PQstatus(R_CONN.conn) == CONNECTION_OK
        ) return(0);
    ifBLIN_QV4(options->flags) {
        fprintf( stderr
               , "Connection defaults:\n"
                 "size=chr keyword	envvar		compiled val		label\n"
               );
        popt = PQconndefaults();
        for (i = 0; popt[i].keyword; i++) {
            fprintf( stderr
                   , "%4d=%s~	%s%s%s%s%s%s%s%s%s\n"
                   , popt[i].dispsize
                   , popt[i].dispchar
                   , ALIGN8(keyword)
                   , ALIGN16(envvar)
                   , PRINTNOTNULL(popt[i].compiled)
                   , !(popt[i].compiled) ? "	" : strlen(popt[i].compiled) < 7 ? "~	" : "~ "
                   , PRINTNOTNULL(popt[i].val)
                   , !(popt[i].val)
                   ? "		"
                   : strlen(popt[i].val) < 7 ? "~		" : "~	"
                   , popt[i].label
                   );
        }
        PQconninfoFree(popt);
            fprintf( stderr
                   , "PGREALM=%s%s PGDATESTYLE=%s%s PGTZ=%s%s PGCLIENTENCODING=%s%s PGGEQO=%s%s\n"
                   , PRINTANDNULL(getenv("PGREALM"))
                   , PRINTANDNULL(getenv("PGDATESTYLE"))
                   , PRINTANDNULL(getenv("PGTZ"))
                   , PRINTANDNULL(getenv("PGCLIENTENCODING"))
                   , PRINTANDNULL(getenv("PGGEQO"))
                   );
    }
    R_CONN.conn = PQsetdbLogin( R_CONN.host, R_CONN.port, NULL, NULL
                              , R_CONN.dbname, R_CONN.username, NULL
                              );
    ifBLIN_QV4(options->flags) {
        fprintf(stderr, "PQoptions=%s~\n", PQoptions(R_CONN.conn));
        fprintf( stderr
               , "PQdb=%s~ PQuser=%s~ PQpass=%s%s "
                 "PQhost=%s%s PQport=%s%s "
                 "PQtty=%s%s\n"
               , PQdb(R_CONN.conn), PQuser(R_CONN.conn), PRINTANDNULL(PQpass(R_CONN.conn))
               , PRINTANDNULL(PQhost(R_CONN.conn)), PRINTANDNULL(PQport(R_CONN.conn))
               , PRINTANDNULL(PQtty(R_CONN.conn))
               );
        fprintf( stderr
               , "server_version=%s%s "
                 "client_encoding=%s%s "
                 "is_superuser=%s%s "
                 "session_authorization=%s%s "
                 "DateStyle=%s%s\n"
               , PQSTATNDNULL("server_version")
               , PQSTATNDNULL("client_encoding")
               , PQSTATNDNULL("is_superuser")
               , PQSTATNDNULL("session_authorization")
               , PQSTATNDNULL("DateStyle")
               );
        fprintf( stderr
               , "PQprotocolVersion=%d PQsocket=%d PQbackendPID=%d\n"
               , PQprotocolVersion(R_CONN.conn)
               , PQsocket(R_CONN.conn)
               , PQbackendPID(R_CONN.conn)
               );
    }
    ifBLIN_QV3(options->flags) fprintf(stderr, "PQstatus %d\n", PQstatus(R_CONN.conn));
    PQsetNoticeProcessor(R_CONN.conn, NoticeProcessor, options);
    if  (PQstatus(R_CONN.conn) != CONNECTION_BAD) {
        if  (!(R_CONN.flags & (PGOBLIN_INT64TIME | PGOBLIN_FLOATTIME))) {
            if  (!R_CONN.intran) {
/*          if  (PQtransactionStatus(R_CONN.conn) != PQTRANS_INTRANS) { */
                PQEXEC(res, "BEGIN", PGRES_COMMAND_OK, -1);
                PQclear(res);
            }
            PQEXEC( res
                  , "DECLARE pgoblin BINARY NO SCROLL CURSOR FOR "
                    "SELECT '2000-01-01 00:00:01'::timestamp;"
                    "FETCH 1 FROM pgoblin"
                  , PGRES_TUPLES_OK, -1
                  );
            if  (  PQgetisnull(res, 0, 0)
                || (PQgetlength(res, 0, 0) < 8)
                ) {
                ex = EX_DATAERR;
                errno = EINVAL;
            } else {
                if  (*((u_char*)(PQgetvalue(res, 0, 0))) == 077) {
                    R_CONN.flags |= PGOBLIN_FLOATTIME;
                    ifBLIN_QV3(options->flags)
                        fprintf(stderr, "conn %d float timestamp format\n", a[PGO_PCON]);
                } else if (*((u_char*)(PQgetvalue(res, 0, 0))) == 0) {
                    R_CONN.flags |= PGOBLIN_INT64TIME;
                    ifBLIN_QV3(options->flags)
                        fprintf(stderr, "conn %d int64 timestamp format\n", a[PGO_PCON]);
                } else {
                    ifBLIN_QV1(options->flags)
                        fprintf(stderr, "conn %d unknown timestamp format\n", a[PGO_PCON]);
                    ex = EX_DATAERR;
                    errno = EFTYPE;
            }   }
            PQclear(res);
            PQEXEC(res, "CLOSE pgoblin", PGRES_COMMAND_OK, -1);
            PQclear(res);
            if  (!R_CONN.intran) {
                PQEXEC(res, "END", PGRES_COMMAND_OK, -1);
                PQclear(res);
        }   }
        return(0);
    }
    ifBLIN_QV1(options->flags)
        warnx("Connection to PostgreSQL failed: %s\n", PQerrorMessage(R_CONN.conn));
    return(1);
}

pgoblin_p *
pgoblin_init(pgoblin_con **options, u_int32_t flags) {
    pgoblin_p *pgm = NULL;

    ifBLIN_QV3(flags) fprintf(stderr, "+pgoblin_init options=%s\n", options ? "OK" : "NULL");
    if  (!*options) {
        *options = malloc(sizeof(pgoblin_con));
        if  (*options) {
            bzero(*options, sizeof(pgoblin_con));
            (*options)->flags = flags;
            strncpy((*options)->vers, VERS, 4);
            (*options)->io[0].flags = PGOBLIN_OUTSET;
            (*options)->io[0].onu = fileno(stdout);
            (*options)->io[0].wri = mife_writ;
            (*options)->io[0].clo = close;
            (*options)->maxtmout = 3600000;
    }   }
    if  (*options) {
        if  (!((*options)->delim[0])) strncpy((*options)->delim, "#", 3);
        if  ((pgm = malloc(sizeof(pgoblin_t) * PGOBLIN_PROGSIZE + sizeof(pgoblin_p)))) {
            pgm->maxx = PGOBLIN_PROGSIZE;
            pgm->curr = 0;
        } else {
            errno = ENOMEM;
        }
    } else {
        errno = ENOMEM;
    }
    ifBLIN_QV3(flags) fprintf(stderr, "-pgoblin_init curr=%d maxx=%d\n", pgm->curr, pgm->maxx);
    return(pgm);
}
